/*******************************************************************************
 * Copyright (c) 2003, 2007 s IT Solutions AT Spardat GmbH .
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     s IT Solutions AT Spardat GmbH - initial API and implementation
 *******************************************************************************/

// @(#) $Id: PageSynchronization.java 10334 2013-02-07 15:01:59Z hoenninger $
package at.spardat.xma.mdl.test;

import junit.framework.TestCase;

import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;

import at.spardat.enterprise.util.Types;
import at.spardat.xma.component.ComponentClient;
import at.spardat.xma.component.ComponentServer;
import at.spardat.xma.mdl.AttachmentExceptionClient;
import at.spardat.xma.mdl.WModel;
import at.spardat.xma.mdl.list.ListDomWM;
import at.spardat.xma.mdl.list.ListDomWMClient;
import at.spardat.xma.mdl.list.ListWM;
import at.spardat.xma.mdl.list.ListWMClient;
import at.spardat.xma.mdl.simple.SimpleWM;
import at.spardat.xma.mdl.simple.SimpleWMClient;
import at.spardat.xma.mdl.table.TableWM;
import at.spardat.xma.mdl.table.TableWMClient;
import at.spardat.xma.mdl.tree.TreeWM;
import at.spardat.xma.mdl.tree.TreeWMClient;
import at.spardat.xma.page.Page;
import at.spardat.xma.page.PageClient;
import at.spardat.xma.page.PageServer;
import at.spardat.xma.page.Scaler;
import at.spardat.xma.serializer.Deserializer;
import at.spardat.xma.serializer.Serializer;
import at.spardat.xma.serializer.SerializerFactoryClient;
import at.spardat.xma.serializer.SerializerFactoryServer;
import at.spardat.xma.serializer.XmaInput;
import at.spardat.xma.serializer.XmaOutput;
import at.spardat.xma.test.TestUtil;


/**
 * Tests the synchronization between two pages
 *
 * @author YSD, 03.05.2003 10:46:16
 */
public class PageSynchronization extends TestCase {

	/**
     * Constructor for PageSynchronization.
     * @param name
     */
    public PageSynchronization(String name) {
        super(name);
    }

    public void testSync () throws Exception {
        for (int i=0; i<10000; i++) {
            makeSyncs (10);
        }
    }

    /**
     * Constructs two pages with random widget models, makes <tt>nSyncs</tt>
     * synchronizations and modifies the models on both side, the client
     * and the server side between each sychronization.
     */
    public void makeSyncs (int nSyncs) throws Exception {
        ComponentClient     cc = new ComponentClient() {
            public String getModelClass (short id) { return "dummy"; }
            public String getResourceString(short id) { return null; }
            public void invoke(Composite parent) {}
        };
        ComponentServer     cs = new ComponentServer() {
            public String getModelClass (short id) { return "dummy"; }
        };
        Page [] pms = newRandomPagePair (cc, cs, 1, 5, (short)47, (short)37);
        Page       pc = pms[0];
        PageServer       ps = (PageServer) pms[1];
        for (int i=0; i<nSyncs; i++) {
            // before synchronizing, we reset stateless server pages or
            // server widget models which are one way
            if (ps.isStateless()) ps.clear();
            else {
                // search for one way tables and remove their rows (not selection)
                WModel [] wModels = ps.getWModels();
                for (int j=0; j<wModels.length; j++) {
                    if (wModels[j] instanceof TableWM) {
                        if (((TableWM)wModels[j]).isOneWay()) ((TableWM)(wModels[j])).internalRemoveRows();
                    }
                }
            }
            SerializerFactoryClient     cfac = new SerializerFactoryClient();
            cfac.setSerializationMode("binary");
            randomlyAddModels((PageClient)pc);
            pc.randomlyChange(TestUtil.randomInt(0, 20));
            Serializer       ser = cfac.createSerializer(null, 512);
            XmaOutput        csOut = ser;
            boolean                csFull = pc.isStateless();
            pc.externalize (csOut, csFull);
            Deserializer     deser = cfac.createDeserializer(null, ser.getResult().getBytes());
            XmaInput         csIn = deser;
            ps.internalize (csIn);
            ps.commit();
            // here is syncpoint 0; the two pages must be equal
            try {
                pc.equalsCS (ps, 0);
            } catch (Exception ex) {
                printPages(pc, ps, "[C->S]");
                throw ex;
            }
            if (false) {
                printPages(pc, ps, "[C->S]");
            }
            pc.commit ();
            // randomly change the server page
            ps.randomlyChange(TestUtil.randomInt(0, 20));
            // rollback changes
            ps.rollback();
            // all changes must be backed out
            try {
                pc.equalsCS (ps, 0);
            } catch (Exception ex) {
                printPages(pc, ps, "after server rollback [C->S]");
                throw ex;
            }
            if (false) {
                printPages(pc, ps, "after server rollback [C->S]");
            }
            // randomly change the server page
            randomlyAddModels((PageServer)ps);
            ps.randomlyChange(TestUtil.randomInt(0, 20));

            SerializerFactoryServer     sfac = new SerializerFactoryServer();
            sfac.setSerializationMode("binary");
            ser = sfac.createSerializer(null, 512);
            XmaOutput        scOut = ser;
            ps.externalize (scOut, false);
            deser = sfac.createDeserializer(null, ser.getResult().getBytes());
            XmaInput         scIn = deser;
            pc.internalize (scIn);
            // here is syncpoint 1; the two pages must be equal
            try {
                pc.equalsCS (ps, 1);
            } catch (Exception ex) {
                printPages(pc, ps, "[S->C]");
                throw ex;
            }
            if (false) {
                printPages(pc, ps, "[S->C]");
            }
            ps.commit ();
        }
    }

    private void printPages (Page pc, PageServer ps, String praefix) {
        // print client page
        System.out.println (praefix + " client page:");
        System.out.println (pc);
        // print server page
        System.out.println (praefix + " server page:");
        System.out.println (ps);
    }

    /**
     * Constructs a pair of pages (a client and a server page)
     *
     * @param cc  the client component
     * @param sc  the server component
     * @param minWidgets min number of widget models
     * @param maxWidgets max number of widget models
     * @param typeId type of the page
     * @param pageId id of the page
     * @return array x where <tt>x[0]</tt> is the client page and <tt>x[1]</tt> is the server page
     */
    public Page[] newRandomPagePair (ComponentClient cc, ComponentServer sc, int minWidgets, int maxWidgets, short typeId, short pageId) {
        boolean stateless = TestUtil.draw(0.5);
        TestPageClient cp = new TestPageClient (cc, stateless, typeId, pageId);
        TestPageServer sp = new TestPageServer (sc, stateless, typeId, pageId);
        WModel[][] wModels = newRandomWModels(cp, sp, (short)0, minWidgets, maxWidgets);
        cp.setWModelsFrom(wModels);
        sp.setWModelsFrom(wModels);
        return new Page[] { cp, sp };
    }


    /**
     * Returns a two column matrix of randomly created widget models where the two
     * entries at a particular index are client- and server buddies. The WModel
     * at index 0 is the client widget model, the one at index 1 the server widget
     * model.
     *
     * @param pc the client page
     * @param ps the server page
     * @param countMin minimum number of widget models to create
     * @param countMax maximun number of widget models
     * @return array with two columns and <tt>k</tt> rows having <tt>[countMin &le; k &le; countMax]</tt>.
     */
    public WModel[][] newRandomWModels (PageClient pc, PageServer ps,short baseID, int countMin, int countMax) {
        int countWM = TestUtil.randomInt (countMin, countMax);
        WModel[][] models = new WModel[countWM][2];
        for (short i=0; i<countWM; i++) {
            switch (TestUtil.randomInt(0, 6)) {
                case 0: newRandomSimpleWM (pc, ps, models[i], (short)(baseID+i)); break;
                case 1: newRandomListAutoWM (pc, ps, models[i], (short)(baseID+i)); break;
                // some more probality for tables because they are more complex
                case 2:
                case 3: newRandomTableDirectWM (pc, ps, models[i], (short)(baseID+i)); break;
                case 4:
                case 5: newRandomTreeDirectWM (pc, ps, models[i], (short)(baseID+i)); break;
                case 6: newRandomListWM(pc,ps,models[i],(short)(baseID+i)); break;
            }
        }
        return models;
    }

    // randomly constructs a SimpleWM pair
    public void newRandomSimpleWM (PageClient pc, PageServer ps, WModel[] models, short id) {
        byte type = (byte) TestUtil.randomInt(Types.FIRST, Types.LAST);
        models[0] = new SimpleWMClient (id, type, pc);
        models[1] = new SimpleWM (id, type, ps);
    }

    // randomly constructs a ListDomWM pair
    public void newRandomListAutoWM (PageClient pc, PageServer ps, WModel[] models, short id) {
        int style = 0;
        if (TestUtil.draw(0.5)) style |= ListDomWMClient.MANDATORY;
        switch (TestUtil.randomInt(0, 1)) {
            case 0: style |= ListDomWMClient.SHOW_LONG; break;
            case 1: style |= ListDomWMClient.SHOW_SHORT_LONG; break;
        }
        switch (TestUtil.randomInt(0, 2)) {
            case 0: style |= ListDomWMClient.ORDER_LONG; break;
            case 1: style |= ListDomWMClient.ORDER_NATURAL; break;
            case 2: style |= ListDomWMClient.ORDER_SHORT; break;
        }
        String dataSource = TestUtil.draw(0.5) ? "Gender" : "StatusXMA";
        ListDomWMClient mc = new ListDomWMClient (id, pc, dataSource, style);
        mc.setOrderStyle(style);
        models[0] = mc;
        models[1] = new ListDomWM (id, ps, dataSource);
    }

    // randomly constructs a TableWM pair
    public void newRandomTableDirectWM (PageClient pc, PageServer ps, WModel[] models, short id) {
        int colCount = TestUtil.randomInt(1, 3);
        int style = 0;
        if (TestUtil.draw(0.5)) style |= TableWM.S_MULTI_SELECT;
        if (TestUtil.draw(0.5)) style |= TableWM.S_ONE_WAY;
        models[0] = new TableWMClient (id, pc, colCount, style);
        models[1] = new TableWM (id, ps, colCount, style);
    }

    // randomly constructs a TreeWM pair
    public void newRandomTreeDirectWM (PageClient pc, PageServer ps, WModel[] models, short id) {
        int style = 0;
        if (TestUtil.draw(0.5)) style |= TreeWM.S_MULTI_SELECT;
        models[0] = new TreeWMClient (id, pc, style);
        models[1] = new TreeWM(id, ps, style);
    }

    // randomly constructs a TableWM pair
    public void newRandomListWM (PageClient pc, PageServer ps, WModel[] models, short id) {
        int style = 0;
        if (TestUtil.draw(0.5)) style |= ListWM.S_MULTI_SELECT;
        if (TestUtil.draw(0.5)) style |= ListWM.S_NOT_STRICT;
        models[0] = new ListWMClient (id, pc, style);
        models[1] = new ListWM (id, ps, style);
    }

    void randomlyAddModels(PageClient pc) {
    	if(TestUtil.draw(0.1)) {
    		WModel[][] newModels = newRandomWModels(pc,null,pc.getNextModelId(),0,TestUtil.randomInt(0,5));
    		for(int i=0;i<newModels.length;i++) {
    			pc.addWModel(newModels[i][0]);
    		}
    	}
    }

    void randomlyAddModels(PageServer ps) {
    	if(TestUtil.draw(0.1)) {
    		WModel[][] newModels = newRandomWModels(null,ps,ps.getNextModelId(),0,TestUtil.randomInt(0,5));
    		for(int i=0;i<newModels.length;i++) {
    			ps.addWModel(newModels[i][1]);
    		}
    	}
    }

    /**
     * Constructs a test page.
     *
     * @author YSD, 03.05.2003 10:49:49
     */
    private static class TestPageServer extends PageServer {

        /**
         * Constructs a page and copies its widgets from <tt>wModels</tt>.
         */
        public TestPageServer (ComponentServer c, boolean stateless, short typeId, short id) {
            super (c, stateless);
            typeId_ = typeId;
            setId(id);
        }

        public void setWModelsFrom (WModel [][] wModels) {
            int col = isAtServer() ? 1 : 0;
            staticModels=wModels.length;
            wModels_ = new WModel[wModels.length];
            for (int i=0; i<wModels.length; i++) wModels_[i] = wModels[i][col];
        }

        protected void setWModels(WModel[] wModels) {
        	wModels_=wModels;
        }

        /**
         * @see at.spardat.xma.mdl.Page#getTypeId()
         */
        public short getTypeId() {
            return typeId_;
        }

        /**
         * @see at.spardat.xma.mdl.Page#getWModels()
         */
        public WModel[] getWModels() {
            return wModels_;
        }

        public void clear() {
        	WModel[] models = new WModel[staticModels];
        	System.arraycopy(wModels_,0,models,0,staticModels);
        	wModels_=models;
        	super.clear();
        }

        private short           typeId_;
        private WModel []       wModels_;
        private int				staticModels;

        public void setChild(PageServer child) { }

        public void removeChild(PageServer child) { }
    }

    /**
     * Constructs a test page.
     *
     * @author YSD, 03.05.2003 10:49:49
     */
    private static class TestPageClient extends PageClient {

        /**
         * Constructs a page and copies its widgets from <tt>wModels</tt>.
         */
        public TestPageClient (ComponentClient c, boolean stateless, short typeId, short id) {
            super (c, stateless);
            typeId_ = typeId;
            setId(id);
        }

        /**
         * @see at.spardat.xma.mdl.PageClient#attachUIImpl()
         */
        public void attachUIImpl() throws AttachmentExceptionClient {
        }


        public void setWModelsFrom (WModel [][] wModels) {
            int col = isAtServer() ? 1 : 0;
            wModels_ = new WModel[wModels.length];
            for (int i=0; i<wModels.length; i++) wModels_[i] = wModels[i][col];
        }

        protected void setWModels(WModel[] wModels) {
        	wModels_=wModels;
        }

        /**
         * @see at.spardat.xma.mdl.Page#getTypeId()
         */
        public short getTypeId() {
            return typeId_;
        }

        /**
         * @see at.spardat.xma.mdl.Page#getWModels()
         */
        public WModel[] getWModels() {
            return wModels_;
        }

        /**
         * @see at.spardat.xma.page.PageClient#createModels()
         */
        protected void createModels() {
        }

        /**
         * @see at.spardat.xma.page.PageClient#createWidgets()
         */
        protected void createWidgets() {
        }

        protected void removeWidgets() {
        }
        
        protected Scaler getScaler () {
            return null;
        }        

        /**
         * @see at.spardat.xma.page.PageClient#hasModels()
         */
        public boolean hasModels() {
            return true;
        }

        /**
         * @see at.spardat.xma.page.PageClient#removeWidgetModels()
         */
        protected void removeWidgetModels() {

        }

        /**
         * @see at.spardat.xma.page.IXMAControl#createComposite(org.eclipse.swt.widgets.Composite)
         */
        public Composite createComposite(Composite parent) {
            return null;
        }
        
        

        /**
         * @see at.spardat.xma.page.PageClient#getWidgets()
         */
        public Control[] getWidgets() {
            return null;
        }

        private short           typeId_;
        private WModel []       wModels_;
    }

}
